@@ -1,7 +1,5 @@
 /*
-    Copyright (c) 2009-2011 250bpm s.r.o.
-    Copyright (c) 2007-2009 iMatix Corporation
-    Copyright (c) 2007-2011 Other contributors as noted in the AUTHORS file
+    Copyright (c) 2007-2013 Contributors as noted in the AUTHORS file
 
     This file is part of 0MQ.
 
@@ -38,19 +36,25 @@
 #include "stream_engine.hpp"
 #include "io_thread.hpp"
 #include "session_base.hpp"
-#include "encoder.hpp"
-#include "decoder.hpp"
 #include "v1_encoder.hpp"
 #include "v1_decoder.hpp"
+#include "v2_encoder.hpp"
+#include "v2_decoder.hpp"
+#include "null_mechanism.hpp"
+#include "plain_mechanism.hpp"
+#include "curve_client.hpp"
+#include "curve_server.hpp"
+#include "raw_decoder.hpp"
+#include "raw_encoder.hpp"
 #include "config.hpp"
 #include "err.hpp"
 #include "ip.hpp"
 #include "likely.hpp"
 #include "wire.hpp"
 
-zmq::stream_engine_t::stream_engine_t (fd_t fd_, const options_t &options_, const std::string &endpoint_) :
+zmq::stream_engine_t::stream_engine_t (fd_t fd_, const options_t &options_, 
+                                       const std::string &endpoint_) :
     s (fd_),
-    io_enabled (false),
     inpos (NULL),
     insize (0),
     decoder (NULL),
@@ -58,40 +62,36 @@ zmq::stream_engine_t::stream_engine_t (fd_t fd_, const options_t &options_, cons
     outsize (0),
     encoder (NULL),
     handshaking (true),
+    greeting_size (v2_greeting_size),
     greeting_bytes_read (0),
     session (NULL),
     options (options_),
     endpoint (endpoint_),
     plugged (false),
+    terminating (false),
+    read_msg (&stream_engine_t::read_identity),
+    write_msg (&stream_engine_t::write_identity),
+    io_error (false),
+    subscription_required (false),
+    mechanism (NULL),
+    input_paused (false),
+    output_paused (false),
     socket (NULL)
 {
+    int rc = tx_msg.init ();
+    errno_assert (rc == 0);
+    
     //  Put the socket into non-blocking mode.
     unblock_socket (s);
-    //  Set the socket buffer limits for the underlying socket.
-    if (options.sndbuf) {
-        int rc = setsockopt (s, SOL_SOCKET, SO_SNDBUF,
-            (char*) &options.sndbuf, sizeof (int));
-#ifdef ZMQ_HAVE_WINDOWS
-		wsa_assert (rc != SOCKET_ERROR);
-#else
-        errno_assert (rc == 0);
-#endif
-    }
-    if (options.rcvbuf) {
-        int rc = setsockopt (s, SOL_SOCKET, SO_RCVBUF,
-            (char*) &options.rcvbuf, sizeof (int));
-#ifdef ZMQ_HAVE_WINDOWS
-		wsa_assert (rc != SOCKET_ERROR);
-#else
-        errno_assert (rc == 0);
-#endif
-    }
+
+    if (!get_peer_ip_address (s, peer_address))
+        peer_address = "";
 
 #ifdef SO_NOSIGPIPE
     //  Make sure that SIGPIPE signal is not generated when writing to a
     //  connection that was already closed by the peer.
     int set = 1;
-    int rc = setsockopt (s, SOL_SOCKET, SO_NOSIGPIPE, &set, sizeof (int));
+    rc = setsockopt (s, SOL_SOCKET, SO_NOSIGPIPE, &set, sizeof (int));
     errno_assert (rc == 0);
 #endif
 }
@@ -102,19 +102,21 @@ zmq::stream_engine_t::~stream_engine_t ()
 
     if (s != retired_fd) {
 #ifdef ZMQ_HAVE_WINDOWS
-		int rc = closesocket (s);
-		wsa_assert (rc != SOCKET_ERROR);
+        int rc = closesocket (s);
+        wsa_assert (rc != SOCKET_ERROR);
 #else
-		int rc = close (s);
+        int rc = close (s);
         errno_assert (rc == 0);
 #endif
-		s = retired_fd;
+        s = retired_fd;
     }
 
-    if (encoder != NULL)
-        delete encoder;
-    if (decoder != NULL)
-        delete decoder;
+    int rc = tx_msg.close ();
+    errno_assert (rc == 0);
+
+    delete encoder;
+    delete decoder;
+    delete mechanism;
 }
 
 void zmq::stream_engine_t::plug (io_thread_t *io_thread_,
@@ -132,15 +134,31 @@ void zmq::stream_engine_t::plug (io_thread_t *io_thread_,
     //  Connect to I/O threads poller object.
     io_object_t::plug (io_thread_);
     handle = add_fd (s);
-    io_enabled = true;
+    io_error = false;
 
-    //  Send the 'length' and 'flags' fields of the identity message.
-    //  The 'length' field is encoded in the long format.
-    outpos = greeting_output_buffer;
-    outpos [outsize++] = 0xff;
-    put_uint64 (&outpos [outsize], options.identity_size + 1);
-    outsize += 8;
-    outpos [outsize++] = 0x7f;
+    if (options.raw_sock) {
+        // no handshaking for raw sock, instantiate raw encoder and decoders
+        encoder = new (std::nothrow) raw_encoder_t (out_batch_size);
+        alloc_assert (encoder);
+
+        decoder = new (std::nothrow) raw_decoder_t (in_batch_size);
+        alloc_assert (decoder);
+
+        // disable handshaking for raw socket
+        handshaking = false;
+
+        read_msg = &stream_engine_t::pull_msg_from_session;
+        write_msg = &stream_engine_t::push_msg_to_session;
+    }
+    else {
+        //  Send the 'length' and 'flags' fields of the identity message.
+        //  The 'length' field is encoded in the long format.
+        outpos = greeting_send;
+        outpos [outsize++] = 0xff;
+        put_uint64 (&outpos [outsize], options.identity_size + 1);
+        outsize += 8;
+        outpos [outsize++] = 0x7f;
+    }
 
     set_pollin (handle);
     set_pollout (handle);
@@ -154,37 +172,43 @@ void zmq::stream_engine_t::unplug ()
     plugged = false;
 
     //  Cancel all fd subscriptions.
-    if (io_enabled) {
+    if (!io_error)
         rm_fd (handle);
-        io_enabled = false;
-    }
 
     //  Disconnect from I/O threads poller object.
     io_object_t::unplug ();
 
-    //  Disconnect from session object.
-    if (encoder)
-        encoder->set_msg_source (NULL);
-    if (decoder)
-        decoder->set_msg_sink (NULL);
     session = NULL;
 }
 
 void zmq::stream_engine_t::terminate ()
 {
+    if (!terminating && encoder && encoder->has_data ()) {
+        //  Give io_thread a chance to send in the buffer
+        terminating = true;
+        return;
+    }
     unplug ();
     delete this;
 }
 
 void zmq::stream_engine_t::in_event ()
 {
-    //  If still handshaking, receive and prcess the greeting message.
+    assert (!io_error);
+
+    //  If still handshaking, receive and process the greeting message.
     if (unlikely (handshaking))
         if (!handshake ())
             return;
 
     zmq_assert (decoder);
-    bool disconnection = false;
+
+    //  If there has been an I/O error, stop polling.
+    if (input_paused) {
+        rm_fd (handle);
+        io_error = true;
+        return;
+    }
 
     //  If there's no data to process in the buffer...
     if (!insize) {
@@ -194,52 +218,51 @@ void zmq::stream_engine_t::in_event ()
         //  the underlying TCP layer has fixed buffer size and thus the
         //  number of bytes read will be always limited.
         decoder->get_buffer (&inpos, &insize);
-        insize = read (inpos, insize);
+        const int bytes_read = read (inpos, insize);
 
         //  Check whether the peer has closed the connection.
-        if (insize == (size_t) -1) {
-            insize = 0;
-            disconnection = true;
+        if (bytes_read == -1) {
+            error ();
+            return;
         }
+
+        //  Adjust input size
+        insize = static_cast <size_t> (bytes_read);
     }
 
-    //  Push the data to the decoder.
-    size_t processed = decoder->process_buffer (inpos, insize);
+    int rc = 0;
+    size_t processed = 0;
 
-    if (unlikely (processed == (size_t) -1)) {
-        disconnection = true;
-    }
-    else {
-
-        //  Stop polling for input if we got stuck.
-        if (processed < insize)
-            reset_pollin (handle);
-
-        //  Adjust the buffer.
+    while (insize > 0) {
+        rc = decoder->decode (inpos, insize, processed);
+        zmq_assert (processed <= insize);
         inpos += processed;
         insize -= processed;
+        if (rc == 0 || rc == -1)
+            break;
+        rc = (this->*write_msg) (decoder->msg ());
+        if (rc == -1)
+            break;
     }
 
-    //  Flush all messages the decoder may have produced.
-    session->flush ();
-
-    //  Input error has occurred. If the last decoded
-    //  message has already been accepted, we terminate
-    //  the engine immediately. Otherwise, we stop
-    //  waiting for input events and postpone the termination
-    //  until after the session has accepted the message.
-    if (disconnection) {
-        if (decoder->stalled ()) {
-            rm_fd (handle);
-            io_enabled = false;
-        }
-        else
+    //  Tear down the connection if we have failed to decode input data
+    //  or the session has rejected the message.
+    if (rc == -1) {
+        if (errno != EAGAIN) {
             error ();
+            return;
+        }
+        input_paused = true;
+        reset_pollin (handle);
     }
+
+    session->flush ();
 }
 
 void zmq::stream_engine_t::out_event ()
 {
+    zmq_assert (!io_error);
+
     //  If write buffer is empty, try to read new data from the encoder.
     if (!outsize) {
 
@@ -252,10 +275,23 @@ void zmq::stream_engine_t::out_event ()
         }
 
         outpos = NULL;
-        encoder->get_data (&outpos, &outsize);
+        outsize = encoder->encode (&outpos, 0);
+
+        while (outsize < out_batch_size) {
+            if ((this->*read_msg) (&tx_msg) == -1)
+                break;
+            encoder->load_msg (&tx_msg);
+            unsigned char *bufptr = outpos + outsize;
+            size_t n = encoder->encode (&bufptr, out_batch_size - outsize);
+            zmq_assert (n > 0);
+            if (outpos == NULL)
+                outpos = bufptr;
+            outsize += n;
+        }
 
         //  If there is no data to send, stop polling for output.
         if (outsize == 0) {
+            output_paused = true;
             reset_pollout (handle);
             return;
         }
@@ -263,16 +299,18 @@ void zmq::stream_engine_t::out_event ()
 
     //  If there are any data to write in write buffer, write as much as
     //  possible to the socket. Note that amount of data to write can be
-    //  arbitratily large. However, we assume that underlying TCP layer has
+    //  arbitrarily large. However, we assume that underlying TCP layer has
     //  limited transmission buffer and thus the actual number of bytes
     //  written should be reasonably modest.
     int nbytes = write (outpos, outsize);
 
     //  IO error has occurred. We stop waiting for output events.
     //  The engine is not terminated until we detect input error;
-    //  this is necessary to prevent losing incomming messages.
+    //  this is necessary to prevent losing incoming messages.
     if (nbytes == -1) {
         reset_pollout (handle);
+        if (unlikely (terminating))
+            terminate ();
         return;
     }
 
@@ -284,11 +322,21 @@ void zmq::stream_engine_t::out_event ()
     if (unlikely (handshaking))
         if (outsize == 0)
             reset_pollout (handle);
+
+    if (unlikely (terminating))
+        if (outsize == 0)
+            terminate ();
 }
 
 void zmq::stream_engine_t::activate_out ()
 {
-    set_pollout (handle);
+    if (unlikely (io_error))
+        return;
+
+    if (likely (output_paused)) {
+        set_pollout (handle);
+        output_paused = false;
+    }
 
     //  Speculative write: The assumption is that at the moment new message
     //  was sent by the user the socket is probably available for writing.
@@ -299,38 +347,59 @@ void zmq::stream_engine_t::activate_out ()
 
 void zmq::stream_engine_t::activate_in ()
 {
-    if (unlikely (!io_enabled)) {
-        //  There was an input error but the engine could not
-        //  be terminated (due to the stalled decoder).
-        //  Flush the pending message and terminate the engine now.
-        zmq_assert (decoder);
-        decoder->process_buffer (inpos, 0);
-        zmq_assert (!decoder->stalled ());
-        session->flush ();
-        error ();
+    zmq_assert (input_paused);
+    zmq_assert (session != NULL);
+    zmq_assert (decoder != NULL);
+
+    int rc = (this->*write_msg) (decoder->msg ());
+    if (rc == -1) {
+        if (errno == EAGAIN)
+            session->flush ();
+        else
+            error ();
         return;
     }
 
-    set_pollin (handle);
+    while (insize > 0) {
+        size_t processed = 0;
+        rc = decoder->decode (inpos, insize, processed);
+        zmq_assert (processed <= insize);
+        inpos += processed;
+        insize -= processed;
+        if (rc == 0 || rc == -1)
+            break;
+        rc = (this->*write_msg) (decoder->msg ());
+        if (rc == -1)
+            break;
+    }
 
-    //  Speculative read.
-    in_event ();
+    if (rc == -1 && errno == EAGAIN)
+        session->flush ();
+    else
+    if (rc == -1 || io_error)
+        error ();
+    else {
+        input_paused = false;
+        set_pollin (handle);
+        session->flush ();
+
+        //  Speculative read.
+        in_event ();
+    }
 }
 
 bool zmq::stream_engine_t::handshake ()
 {
     zmq_assert (handshaking);
     zmq_assert (greeting_bytes_read < greeting_size);
-
     //  Receive the greeting.
     while (greeting_bytes_read < greeting_size) {
-        const int n = read (greeting + greeting_bytes_read,
+        const int n = read (greeting_recv + greeting_bytes_read,
                             greeting_size - greeting_bytes_read);
         if (n == -1) {
             error ();
             return false;
         }
-
         if (n == 0)
             return false;
 
@@ -339,43 +408,71 @@ bool zmq::stream_engine_t::handshake ()
         //  We have received at least one byte from the peer.
         //  If the first byte is not 0xff, we know that the
         //  peer is using unversioned protocol.
-        if (greeting [0] != 0xff)
+        if (greeting_recv [0] != 0xff)
             break;
 
-        if (greeting_bytes_read < 10)
+        if (greeting_bytes_read < signature_size)
             continue;
 
         //  Inspect the right-most bit of the 10th byte (which coincides
         //  with the 'flags' field if a regular message was sent).
         //  Zero indicates this is a header of identity message
         //  (i.e. the peer is using the unversioned protocol).
-        if (!(greeting [9] & 0x01))
+        if (!(greeting_recv [9] & 0x01))
             break;
 
         //  The peer is using versioned protocol.
-        //  Send the rest of the greeting, if necessary.
-        if (outpos + outsize != greeting_output_buffer + greeting_size) {
+        //  Send the major version number.
+        if (outpos + outsize == greeting_send + signature_size) {
             if (outsize == 0)
                 set_pollout (handle);
-            outpos [outsize++] = 1;             // Protocol version
-            outpos [outsize++] = options.type;  // Socket type
+            outpos [outsize++] = 3;     //  Major version number
+        }
+
+        if (greeting_bytes_read > signature_size) {
+            if (outpos + outsize == greeting_send + signature_size + 1) {
+                if (outsize == 0)
+                    set_pollout (handle);
+
+                //  Use ZMTP/2.0 to talk to older peers.
+                if (greeting_recv [10] == ZMTP_1_0
+                ||  greeting_recv [10] == ZMTP_2_0)
+                    outpos [outsize++] = options.type;
+                else {
+                    outpos [outsize++] = 0; //  Minor version number
+                    memset (outpos + outsize, 0, 20);
+
+                    zmq_assert (options.mechanism == ZMQ_NULL
+                            ||  options.mechanism == ZMQ_PLAIN
+                            ||  options.mechanism == ZMQ_CURVE);
+
+                    if (options.mechanism == ZMQ_NULL)
+                        memcpy (outpos + outsize, "NULL", 4);
+                    else
+                    if (options.mechanism == ZMQ_PLAIN)
+                        memcpy (outpos + outsize, "PLAIN", 5);
+                    else
+                        memcpy (outpos + outsize, "CURVE", 5);
+                    outsize += 20;
+                    memset (outpos + outsize, 0, 32);
+                    outsize += 32;
+                    greeting_size = v3_greeting_size;
+                }
+            }
         }
     }
 
-    //  Position of the version field in the greeting.
-    const size_t version_pos = 10;
+    //  Position of the revision field in the greeting.
+    const size_t revision_pos = 10;
 
-    //  Is the peer using the unversioned protocol?
-    //  If so, we send and receive rests of identity
-    //  messages.
-    if (greeting [0] != 0xff || !(greeting [9] & 0x01)) {
-        encoder = new (std::nothrow) encoder_t (out_batch_size);
+    //  Is the peer using ZMTP/1.0 with no revision number?
+    //  If so, we send and receive rest of identity message
+    if (greeting_recv [0] != 0xff || !(greeting_recv [9] & 0x01)) {
+        encoder = new (std::nothrow) v1_encoder_t (out_batch_size);
         alloc_assert (encoder);
-        encoder->set_msg_source (session);
 
-        decoder = new (std::nothrow) decoder_t (in_batch_size, options.maxmsgsize);
+        decoder = new (std::nothrow) v1_decoder_t (in_batch_size, options.maxmsgsize);
         alloc_assert (decoder);
-        decoder->set_msg_sink (session);
 
         //  We have already sent the message header.
         //  Since there is no way to tell the encoder to
@@ -383,41 +480,74 @@ bool zmq::stream_engine_t::handshake ()
         //  header data away.
         const size_t header_size = options.identity_size + 1 >= 255 ? 10 : 2;
         unsigned char tmp [10], *bufferp = tmp;
-        size_t buffer_size = header_size;
-        encoder->get_data (&bufferp, &buffer_size);
+        size_t buffer_size = encoder->encode (&bufferp, header_size);
         zmq_assert (buffer_size == header_size);
 
         //  Make sure the decoder sees the data we have already received.
-        inpos = greeting;
+        inpos = greeting_recv;
         insize = greeting_bytes_read;
 
         //  To allow for interoperability with peers that do not forward
-        //  their subscriptions, we inject a phony subsription
-        //  message into the incomming message stream. To put this
-        //  message right after the identity message, we temporarily
-        //  divert the message stream from session to ourselves.
+        //  their subscriptions, we inject a phantom subscription message
+        //  message into the incoming message stream.
         if (options.type == ZMQ_PUB || options.type == ZMQ_XPUB)
-            decoder->set_msg_sink (this);
+            subscription_required = true;
     }
     else
-    if (greeting [version_pos] == 0) {
-        //  ZMTP/1.0 framing.
-        encoder = new (std::nothrow) encoder_t (out_batch_size);
+    if (greeting_recv [revision_pos] == ZMTP_1_0) {
+        encoder = new (std::nothrow) v1_encoder_t (
+            out_batch_size);
         alloc_assert (encoder);
-        encoder->set_msg_source (session);
 
-        decoder = new (std::nothrow) decoder_t (in_batch_size, options.maxmsgsize);
+        decoder = new (std::nothrow) v1_decoder_t (
+            in_batch_size, options.maxmsgsize);
+        alloc_assert (decoder);
+    }
+    else
+    if (greeting_recv [revision_pos] == ZMTP_2_0) {
+        encoder = new (std::nothrow) v2_encoder_t (out_batch_size);
+        alloc_assert (encoder);
+
+        decoder = new (std::nothrow) v2_decoder_t (
+            in_batch_size, options.maxmsgsize);
         alloc_assert (decoder);
-        decoder->set_msg_sink (session);
     }
     else {
-        //  v1 framing protocol.
-        encoder = new (std::nothrow) v1_encoder_t (out_batch_size, session);
+        encoder = new (std::nothrow) v2_encoder_t (out_batch_size);
         alloc_assert (encoder);
 
-        decoder = new (std::nothrow)
-            v1_decoder_t (in_batch_size, options.maxmsgsize, session);
+        decoder = new (std::nothrow) v2_decoder_t (
+            in_batch_size, options.maxmsgsize);
         alloc_assert (decoder);
+
+        if (memcmp (greeting_recv + 12, "NULL\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 20) == 0) {
+            mechanism = new (std::nothrow)
+                null_mechanism_t (session, peer_address, options);
+            alloc_assert (mechanism);
+        }
+        else
+        if (memcmp (greeting_recv + 12, "PLAIN\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 20) == 0) {
+            mechanism = new (std::nothrow)
+                plain_mechanism_t (session, peer_address, options);
+            alloc_assert (mechanism);
+        }
+#ifdef HAVE_LIBSODIUM
+        else
+        if (memcmp (greeting_recv + 12, "CURVE\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 20) == 0) {
+            if (options.as_server)
+                mechanism = new (std::nothrow)
+                    curve_server_t (session, peer_address, options);
+            else
+                mechanism = new (std::nothrow) curve_client_t (options);
+            alloc_assert (mechanism);
+        }
+#endif
+        else {
+            error ();
+            return false;
+        }
+        read_msg = &stream_engine_t::next_handshake_command;
+        write_msg = &stream_engine_t::process_handshake_command;
     }
 
     // Start polling for output if necessary.
@@ -431,35 +561,166 @@ bool zmq::stream_engine_t::handshake ()
     return true;
 }
 
-int zmq::stream_engine_t::push_msg (msg_t *msg_)
+int zmq::stream_engine_t::read_identity (msg_t *msg_)
 {
-    zmq_assert (options.type == ZMQ_PUB || options.type == ZMQ_XPUB);
-
-    //  The first message is identity.
-    //  Let the session process it.
-    int rc = session->push_msg (msg_);
+    int rc = msg_->init_size (options.identity_size);
     errno_assert (rc == 0);
+    if (options.identity_size > 0)
+        memcpy (msg_->data (), options.identity, options.identity_size);
+    read_msg = &stream_engine_t::pull_msg_from_session;
+    return 0;
+}
 
-    //  Inject the subscription message so that the ZMQ 2.x peer
-    //  receives our messages.
-    rc = msg_->init_size (1);
-    errno_assert (rc == 0);
-    *(unsigned char*) msg_->data () = 1;
-    rc = session->push_msg (msg_);
-    session->flush ();
+int zmq::stream_engine_t::write_identity (msg_t *msg_)
+{
+    if (options.recv_identity) {
+        msg_->set_flags (msg_t::identity);
+        int rc = session->push_msg (msg_);
+        errno_assert (rc == 0);
+    }
+    else {
+        int rc = msg_->close ();
+        errno_assert (rc == 0);
+        rc = msg_->init ();
+        errno_assert (rc == 0);
+    }
 
-    //  Once we have injected the subscription message, we can
-    //  Divert the message flow back to the session.
-    zmq_assert (decoder);
-    decoder->set_msg_sink (session);
+    if (subscription_required)
+        write_msg = &stream_engine_t::write_subscription_msg;
+    else
+        write_msg = &stream_engine_t::push_msg_to_session;
+
+    return 0;
+}
+
+int zmq::stream_engine_t::next_handshake_command (msg_t *msg_)
+{
+    zmq_assert (mechanism != NULL);
+
+    const int rc = mechanism->next_handshake_command (msg_);
+    if (rc == 0) {
+        msg_->set_flags (msg_t::command);
+        if (mechanism->is_handshake_complete ())
+            mechanism_ready ();
+    }
 
     return rc;
 }
 
+int zmq::stream_engine_t::process_handshake_command (msg_t *msg_)
+{
+    zmq_assert (mechanism != NULL);
+    const int rc = mechanism->process_handshake_command (msg_);
+    if (rc == 0) {
+        if (mechanism->is_handshake_complete ())
+            mechanism_ready ();
+        if (output_paused)
+            activate_out ();
+    }
+
+    return rc;
+}
+
+void zmq::stream_engine_t::zap_msg_available ()
+{
+    zmq_assert (mechanism != NULL);
+
+    const int rc = mechanism->zap_msg_available ();
+    if (rc == -1) {
+        error ();
+        return;
+    }
+    if (input_paused)
+        activate_in ();
+    if (output_paused)
+        activate_out ();
+}
+
+void zmq::stream_engine_t::mechanism_ready ()
+{
+    if (options.recv_identity) {
+        msg_t identity;
+        mechanism->peer_identity (&identity);
+        const int rc = session->push_msg (&identity);
+        if (rc == -1 && errno == EAGAIN) {
+            // If the write is failing at this stage with
+            // an EAGAIN the pipe must be being shut down,
+            // so we can just bail out of the identity set.
+            return;
+        }
+        errno_assert (rc == 0);
+        session->flush ();
+    }
+
+    read_msg = &stream_engine_t::pull_and_encode;
+    write_msg = &stream_engine_t::decode_and_push;
+}
+
+int zmq::stream_engine_t::pull_msg_from_session (msg_t *msg_)
+{
+    return session->pull_msg (msg_);
+}
+
+int zmq::stream_engine_t::push_msg_to_session (msg_t *msg_)
+{
+    return session->push_msg (msg_);
+}
+
+int zmq::stream_engine_t::pull_and_encode (msg_t *msg_)
+{
+    zmq_assert (mechanism != NULL);
+
+    if (session->pull_msg (msg_) == -1)
+        return -1;
+    if (mechanism->encode (msg_) == -1)
+        return -1;
+    return 0;
+}
+
+int zmq::stream_engine_t::decode_and_push (msg_t *msg_)
+{
+    zmq_assert (mechanism != NULL);
+
+    if (mechanism->decode (msg_) == -1)
+        return -1;
+    if (session->push_msg (msg_) == -1) {
+        if (errno == EAGAIN)
+            write_msg = &stream_engine_t::push_one_then_decode_and_push;
+        return -1;
+    }
+    return 0;
+}
+
+int zmq::stream_engine_t::push_one_then_decode_and_push (msg_t *msg_)
+{
+    const int rc = session->push_msg (msg_);
+    if (rc == 0)
+        write_msg = &stream_engine_t::decode_and_push;
+    return rc;
+}
+
+int zmq::stream_engine_t::write_subscription_msg (msg_t *msg_)
+{
+    msg_t subscription;
+
+    //  Inject the subscription message, so that also
+    //  ZMQ 2.x peers receive published messages.
+    int rc = subscription.init_size (1);
+    errno_assert (rc == 0);
+    *(unsigned char*) subscription.data () = 1;
+    rc = session->push_msg (&subscription);
+    if (rc == -1)
+       return -1;
+
+    write_msg = &stream_engine_t::push_msg_to_session;
+    return push_msg_to_session (msg_);
+}
+
 void zmq::stream_engine_t::error ()
 {
     zmq_assert (session);
     socket->event_disconnected (endpoint, s);
+    session->flush ();
     session->detach ();
     unplug ();
     delete this;
@@ -475,7 +736,7 @@ int zmq::stream_engine_t::write (const void *data_, size_t size_)
     //  we'll get an error (this may happen during the speculative write).
     if (nbytes == SOCKET_ERROR && WSAGetLastError () == WSAEWOULDBLOCK)
         return 0;
-		
+        
     //  Signalise peer failure.
     if (nbytes == SOCKET_ERROR && (
           WSAGetLastError () == WSAENETDOWN ||
@@ -515,7 +776,7 @@ int zmq::stream_engine_t::write (const void *data_, size_t size_)
         return -1;
     }
 
-    return (size_t) nbytes;
+    return static_cast <int> (nbytes);
 
 #endif
 }
@@ -575,8 +836,7 @@ int zmq::stream_engine_t::read (void *data_, size_t size_)
     if (nbytes == 0)
         return -1;
 
-    return (size_t) nbytes;
+    return static_cast <int> (nbytes);
 
 #endif
 }
-
